from unittest.mock import call, patch

from arkindex.documents.tasks import selection_worker_results_delete
from arkindex.process.models import Worker, WorkerVersion
from arkindex.project.tests import FixtureTestCase
from arkindex.training.models import Model, ModelVersionState


class TestDeleteSelectionWorkerResults(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page1 = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.page2 = cls.corpus.elements.get(name='Volume 1, page 1v')
        cls.page3 = cls.corpus.elements.get(name='Volume 1, page 2r')
        cls.version = WorkerVersion.objects.first()
        cls.model = Model.objects.create(name='Generic model', public=False)
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            tag='Test',
            hash='A' * 32,
            archive_hash='42',
            size=1337,
        )
        cls.configuration = Worker.objects.get(slug='dla').configurations.create(
            name="DLA configuration", configuration={"value": "test"}
        )

    def test_no_rq_job(self):
        with self.assertRaises(AssertionError) as ctx:
            selection_worker_results_delete(corpus_id=self.corpus.id)
        self.assertEqual(str(ctx.exception), 'This task can only be run in a RQ job context.')

    @patch('arkindex.documents.tasks.get_current_job')
    def test_no_user_id(self, job_mock):
        job_mock.return_value.user_id = None
        with self.assertRaises(AssertionError) as ctx:
            selection_worker_results_delete(corpus_id=self.corpus.id)
        self.assertEqual(str(ctx.exception), 'This task requires a user ID to be defined on the RQ job.')

    @patch('arkindex.documents.tasks.get_current_job')
    @patch('arkindex.documents.tasks.worker_results_delete')
    def test_run(self, worker_results_delete_mock, job_mock):
        self.user.selected_elements.set([self.page1, self.page2])
        # Another user's selection should not be used
        self.superuser.selected_elements.set([self.page3])
        job_mock.return_value.user_id = self.user.id

        selection_worker_results_delete(corpus_id=self.corpus.id)

        self.assertEqual(job_mock.call_count, 1)
        self.assertEqual(job_mock().set_progress.call_count, 2)
        self.assertListEqual(job_mock().set_progress.call_args_list, [
            call(.0),
            call(.5),
        ])

        self.assertEqual(worker_results_delete_mock.call_count, 2)
        self.assertCountEqual(worker_results_delete_mock.call_args_list, [
            call(
                corpus_id=self.corpus.id,
                version_id=None,
                model_version_id=None,
                configuration_id=None,
                element_id=self.page1.id,
            ),
            call(
                corpus_id=self.corpus.id,
                version_id=None,
                model_version_id=None,
                configuration_id=None,
                element_id=self.page2.id,
            ),
        ])

    @patch('arkindex.documents.tasks.get_current_job')
    @patch('arkindex.documents.tasks.worker_results_delete')
    def test_run_filters(self, worker_results_delete_mock, job_mock):
        self.user.selected_elements.set([self.page1, self.page2])
        self.superuser.selected_elements.set([self.page3])
        job_mock.return_value.user_id = self.user.id

        selection_worker_results_delete(
            corpus_id=self.corpus.id,
            version_id=self.version.id,
            model_version_id=self.model_version.id,
            configuration_id=self.configuration.id,
        )

        self.assertEqual(job_mock.call_count, 1)
        self.assertEqual(job_mock().set_progress.call_count, 2)
        self.assertListEqual(job_mock().set_progress.call_args_list, [
            call(.0),
            call(.5),
        ])

        self.assertEqual(worker_results_delete_mock.call_count, 2)
        self.assertCountEqual(worker_results_delete_mock.call_args_list, [
            call(
                corpus_id=self.corpus.id,
                version_id=self.version.id,
                model_version_id=self.model_version.id,
                configuration_id=self.configuration.id,
                element_id=self.page1.id,
            ),
            call(
                corpus_id=self.corpus.id,
                version_id=self.version.id,
                model_version_id=self.model_version.id,
                configuration_id=self.configuration.id,
                element_id=self.page2.id,
            ),
        ])
