from uuid import uuid4

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, TextOrientation, Transcription
from arkindex.process.models import ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestTranscriptionCreate(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.line = cls.corpus.elements.filter(type__slug='text_line').first()
        cls.private_corpus = Corpus.objects.create(name='Private')
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        # Create an user with a read right only on the private corpus
        cls.private_read_user = User.objects.create_user('a@bc.de', 'a')
        cls.private_read_user.verified_email = True
        cls.private_read_user.save()
        cls.private_corpus.memberships.create(user=cls.private_read_user, level=Role.Guest.value)

        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_run = cls.worker_version.worker_runs.filter(process__mode=ProcessMode.Workers).get()
        cls.local_worker_run = cls.worker_version.worker_runs.filter(process__mode=ProcessMode.Local).get()

    def test_require_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:transcription-create', kwargs={'pk': self.line.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Authentication credentials were not provided.'
        })

    def test_write_right(self):
        self.client.force_login(self.private_read_user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.private_page.id}),
                format='json',
                data={'text': 'NEKUDOTAYIM'}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': "A write access to the element's corpus is required."
        })

    def test_no_read_right(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.private_page.id}),
                format='json',
                data={'text': 'NEKUDOTAYIM'}
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_no_element(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': uuid4()}),
                format='json',
                data={'text': 'NEKUDOTAYIM'}
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_manual(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={'text': 'A perfect day in a perfect place'}
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'confidence': None,
            'text': 'A perfect day in a perfect place',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': None,
            'worker_run': None,
        })

        new_ts = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertIsNone(new_ts.confidence)
        self.assertEqual(new_ts.worker_version, None)
        self.assertTrue(self.line.transcriptions.filter(pk=new_ts.id).exists())

    def test_duplicated(self):
        """
        Check the view creates a new manual transcriptions with a similar text and element
        """
        self.client.force_login(self.user)
        ts = self.line.transcriptions.create(text='GLOUBIBOULGA')
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': ts.text,
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(
            Transcription.objects.filter(element=self.line, text=ts.text, worker_version__isnull=True).count(),
            2
        )

    def test_with_orientation(self):
        """
        Check that a transcription is created with the specified orientation
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={'text': 'A perfect day in a perfect place', 'orientation': 'vertical-lr'}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'confidence': None,
            'text': 'A perfect day in a perfect place',
            'orientation': 'vertical-lr',
            'worker_version_id': None,
            'worker_run': None,
        })
        new_ts = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertEqual(new_ts.orientation, TextOrientation.VerticalLeftToRight)

    def test_invalid_orientation(self):
        """
        Specifying an invalid text-orientation causes an error
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={'text': 'A perfect day in a perfect place', 'orientation': 'wiggly'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_no_search(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={'text': 'A classy text line'}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_worker_version(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'NEKUDOTAYIM',
                    'worker_version': str(self.worker_version.id),
                    'confidence': .42
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_version': ['This field is forbidden.'],
        })

    def test_worker_run_local(self):
        """
        A regular user can create a classification with a WorkerRun of their own local process
        """
        self.client.force_login(self.superuser)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'NEKUDOTAYIM',
                    'worker_run_id': str(self.local_worker_run.id),
                    'confidence': .42,
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        tr = Transcription.objects.get(text='NEKUDOTAYIM')
        self.assertEqual(tr.worker_run, self.local_worker_run)
        self.assertEqual(tr.worker_version, self.worker_version)
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'confidence': .42,
            'text': 'NEKUDOTAYIM',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': str(self.worker_version.id),
            'worker_run': {
                "id": str(self.local_worker_run.id),
                "summary": self.local_worker_run.summary,
            },
        })

    def test_worker_run_local_task_auth(self):
        """
        A transcription can be created with a WorkerRun of a Local process
        even when authenticated as a Ponos task from a different process
        """
        # Take the user's local process, so that the user's rights will take over the Ponos task auth
        local_process = self.worker_run.process.creator.processes.get(mode=ProcessMode.Local)
        local_worker_run = local_process.worker_runs.get()
        self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'NEKUDOTAYIM',
                    'worker_run_id': str(local_worker_run.id),
                    'confidence': .42,
                },
                HTTP_AUTHORIZATION=f'Ponos {task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        tr = Transcription.objects.get(text='NEKUDOTAYIM')
        self.assertEqual(tr.worker_run, local_worker_run)
        self.assertEqual(tr.worker_version, local_worker_run.version)
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'confidence': .42,
            'text': 'NEKUDOTAYIM',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': str(local_worker_run.version_id),
            'worker_run': {
                "id": str(local_worker_run.id),
                "summary": local_worker_run.summary,
            },
        })

    def test_worker_run_task(self):
        """
        A transcription can be created with a WorkerRun of a non-local process
        when authenticated as a Ponos task of this process
        """
        self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'NEKUDOTAYIM',
                    'worker_run_id': str(self.worker_run.id),
                    'confidence': .42,
                },
                HTTP_AUTHORIZATION=f'Ponos {task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        tr = Transcription.objects.get(text='NEKUDOTAYIM')
        self.assertEqual(tr.worker_run, self.worker_run)
        self.assertEqual(tr.worker_version, self.worker_version)
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'confidence': .42,
            'text': 'NEKUDOTAYIM',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': str(self.worker_version.id),
            'worker_run': {
                "id": str(self.worker_run.id),
                "summary": self.worker_run.summary,
            },
        })

    def test_worker_run_non_local(self):
        """
        A regular user cannot create a transcription with a WorkerRun of a non-local process
        """
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'NEKUDOTAYIM',
                    'worker_run_id': str(self.worker_run.id),
                    'confidence': .42,
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_run_id': ["Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."],
        })

    def test_worker_run_other_user(self):
        """
        A regular user cannot create a transcription with a WorkerRun of someone else's local process
        """
        worker_run = self.user.processes.get(mode=ProcessMode.Local).worker_runs.first()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'NEKUDOTAYIM',
                    'worker_run_id': str(worker_run.id),
                    'confidence': .42,
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_run_id': ["Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."],
        })

    def test_worker_run_other_process(self):
        """
        A Ponos task cannot create a transcription with a WorkerRun of another process
        """
        # Create a process on the same user, to check that being the process creator has no effect
        process2 = self.worker_run.process.creator.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
        )
        other_worker_run = process2.worker_runs.create(version=self.worker_run.version, parents=[])
        self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'NEKUDOTAYIM',
                    'worker_run_id': str(other_worker_run.id),
                    'confidence': .42,
                },
                HTTP_AUTHORIZATION=f'Ponos {task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_run_id': ["Only the WorkerRuns of the authenticated task's process may be used."],
        })

    def test_worker_run_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'NEKUDOTAYIM',
                    'worker_run_id': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                    'confidence': .42
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_run_id': ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.'],
        })

    def test_worker_run_required_confidence(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'CIRCLE',
                    'worker_run_id': str(self.local_worker_run.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'non_field_errors': [
            'The confidence field must be defined for a transcription with a worker run.'
        ]})

    def test_worker_version_xor_worker_run(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'CIRCLE',
                    'worker_run_id': str(self.local_worker_run.id),
                    'worker_version': str(self.worker_version.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_version': ['This field is forbidden.'],
        })

    def test_element_no_zone(self):
        """
        Transcriptions can also be created on elements without zones
        """
        null_zone_page = self.corpus.elements.create(type=self.page.type)

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': null_zone_page.id}),
                format='json',
                data={
                    'text': 'ELLIPSE',
                    'confidence': .42
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_ts = null_zone_page.transcriptions.get()
        self.assertDictEqual(response.json(), {
            'id': str(new_ts.id),
            'confidence': .42,
            'text': 'ELLIPSE',
            'orientation': 'horizontal-lr',
            'worker_version_id': None,
            'worker_run': None,
        })

    def test_ignore_type(self):
        """
        An old transcription type can still be sent but will be ignored
        """
        self.client.force_login(self.superuser)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': 'CIRCLE',
                    'type': 'paragraph',
                    'confidence': 0.654816358,
                    'worker_run_id': str(self.local_worker_run.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        transcription = self.line.transcriptions.get(text='CIRCLE')
        self.assertEqual(transcription.confidence, 0.654816358)
        self.assertEqual(transcription.worker_version, self.worker_version)
        self.assertEqual(transcription.worker_run, self.local_worker_run)
