from django import forms
from django.contrib import admin, messages
from django.core.exceptions import ValidationError
from enumfields.admin import EnumFieldListFilter

from arkindex.ponos.keys import gen_nonce
from arkindex.ponos.models import FINAL_STATES, GPU, Agent, Artifact, Farm, Secret, Task, encrypt


class ArtifactInline(admin.TabularInline):
    model = Artifact
    fields = ("id", "path", "size", "content_type", "created")
    readonly_fields = fields
    # Artifacts cannot be added
    extra = 0

    def has_change_permission(self, request, obj=None):
        """
        Artifacts cannot be modified since all fields are read-only.
        We have to make this explicit to Django. Otherwise, it tries to update all the artifacts
        without receiving any field from the form when saving, so all the required fields are missing
        and users get "Please correct the error below:" with no actual error being displayed.
        """
        return False


class TaskAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "slug",
        "run",
        "state",
        "process_id",
        "updated",
        "agent",
        "priority",
    )
    list_filter = [("state", EnumFieldListFilter), "agent"]
    list_select_related = ("agent",)
    inlines = [
        ArtifactInline,
    ]
    readonly_fields = (
        "id",
        "created",
        "updated",
        "container",
        "shm_size",
    )
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "id",
                    "slug",
                    "run",
                    "depth",
                    "state",
                    "process",
                    "agent",
                    "priority",
                ),
            },
        ),
        ("GPU", {"fields": ("requires_gpu", "gpu")}),
        ("Dates", {"fields": ("created", "updated", "expiry")}),
        (
            "Docker",
            {
                "fields": (
                    "image",
                    "command",
                    "env",
                    "container",
                    "shm_size",
                    "extra_files",
                )
            },
        ),
    )


class TaskInline(admin.TabularInline):
    model = Task
    fields = ("id", "slug", "run", "depth", "state", "container")
    readonly_fields = ("slug", "run", "depth", "container")
    extra = 0


class GPUInline(admin.TabularInline):
    model = GPU
    fields = ("id", "name", "index", "ram_total")
    readonly_fields = fields
    extra = 0


class AgentAdmin(admin.ModelAdmin):
    model = Agent
    list_display = ("id", "hostname")
    inlines = (GPUInline,)
    readonly_fields = (
        "id",
        "created",
        "updated",
        "last_ping",
        # Use custom admin fields to format total RAM and CPU max frequency
        "ram_total_human",
        "cpu_frequency_human",
        "cpu_load",
        "ram_load",
        "cpu_cores",
        "public_key",
    )
    fieldsets = (
        (
            None,
            {"fields": ("id", "farm", "hostname")},
        ),
        ("Dates", {"fields": ("created", "updated", "last_ping")}),
        ("Auth", {"fields": ("public_key",)}),
        (
            "Hardware",
            {
                "fields": (
                    "ram_total_human",
                    "cpu_frequency_human",
                    "cpu_cores",
                    "cpu_load",
                    "ram_load",
                )
            },
        ),
    )

    def has_add_permission(self, request):
        return False

    def ram_total_human(self, instance):
        """Returns total amount of RAM expressed in GiB"""
        return "{:.1f} GiB".format((instance.ram_total or 0) / (1024**3))

    def cpu_frequency_human(self, instance):
        """Returns CPU max frequency expressed in GHz"""
        return "{:.1f} GHz".format((instance.cpu_frequency or 0) / 1e9)

    # Overrides both admin deletion methods to show error messages when agents
    # have tasks in non-final states. This will display a "deleted successfully"
    # message along the errors, because the Django admin does not make it easy to
    # remove this message, but nothing gets actually deleted.

    def delete_model(self, request, agent):
        try:
            super().delete_model(request, agent)
        except ValidationError as e:
            messages.error(request, e.message)

    def delete_queryset(self, request, queryset):
        hostnames = (
            Task.objects.filter(agent__in=queryset)
            .exclude(state__in=FINAL_STATES)
            .values_list("agent__hostname", flat=True)
            .distinct()
        )
        if hostnames:
            messages.error(
                request,
                "The following agents have tasks in non-final states and cannot be deleted: {}".format(
                    ", ".join(list(hostnames))
                ),
            )
        else:
            super().delete_queryset(request, queryset)


class FarmAdmin(admin.ModelAdmin):
    model = Farm
    list_display = ("id", "name")
    fields = ("id", "name", "seed")
    readonly_fields = ("id",)


class ClearTextSecretForm(forms.ModelForm):
    """
    Allow an administrator to edit a secret content as a cleartext value
    A nonce is generated for newly created secrets
    """

    content = forms.CharField(widget=forms.Textarea, required=True)

    def __init__(self, *args, **kwargs):
        self.instance = kwargs.get("instance")
        # Set initial decrypted value
        if self.instance is not None:
            if "initial" not in kwargs:
                kwargs["initial"] = {}
            kwargs["initial"]["content"] = self.instance.decrypt()

        super().__init__(*args, **kwargs)

    def clean_name(self):
        # Check that name is not already used
        secrets = Secret.objects.all()
        if self.instance:
            secrets = Secret.objects.exclude(pk=self.instance.pk)

        if secrets.filter(name=self.cleaned_data["name"]).exists():
            raise ValidationError("A secret with this name already exists.")

        return self.cleaned_data["name"]

    def clean(self):
        if not self.cleaned_data.get("content"):
            raise ValidationError("You must specify some content")

        if self.instance:
            nonce = self.instance.nonce
        else:
            nonce = gen_nonce()

        # Encrypt secret content
        encrypted_content = encrypt(nonce, self.cleaned_data["content"])
        return {**self.cleaned_data, "nonce": nonce, "content": encrypted_content}

    class Meta:
        model = Secret
        fields = ("id", "name", "content")


class SecretAdmin(admin.ModelAdmin):
    form = ClearTextSecretForm

    fields = ("name", "content")


admin.site.register(Task, TaskAdmin)
admin.site.register(Agent, AgentAdmin)
admin.site.register(Farm, FarmAdmin)
admin.site.register(Secret, SecretAdmin)
