from unittest.mock import patch

from django.core import mail
from django.urls import reverse
from rest_framework import status

from arkindex.process.providers import GitLabProvider, GitProvider
from arkindex.project.tests import FixtureAPITestCase

MAIL_CONTENT = """
Hello Test user,

An error occurred while processing a new push on your repository http://my_repo.fake/workers/worker.

Error: Error during handle_webhook

No process will be launched to import your new revision.
Please, try again with another push or contact your system administrator.

--
Arkindex

"""

REFRESH_TOKEN_MAIL_CONTENT = """
Hello Test user,

An error occurred while processing a new push on your repository http://my_repo.fake/workers/worker.

Error: The OAuth token could not be refreshed. Please reconnect the OAuth account manually.
Original error: No refresh token was stored for this OAuth token.

No process will be launched to import your new revision.
Please, try again with another push or contact your system administrator.

--
Arkindex

"""


class TestProviders(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(url='http://my_repo.fake/workers/worker')
        cls.rev = cls.repo.revisions.get()

    def test_init(self):
        glp = GitLabProvider()
        self.assertIsNone(glp.credentials)

        glp = GitLabProvider(credentials=self.creds)
        self.assertEqual(glp.credentials, self.creds)

        with self.assertRaises(Exception):
            GitLabProvider(credentials='not a OAuthCredentials')

    @patch('arkindex.process.api.Repository.provider_class')
    def test_webhook_no_credentials(self, provider_class):
        self.client.force_login(self.user)
        self.repo.credentials = None
        self.repo.save()
        response = self.client.post(reverse('api:import-hook', kwargs={'pk': self.repo.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'No credentials available for this repository.'})
        self.assertEqual(len(mail.outbox), 0)

    @patch('arkindex.process.api.Repository.provider_class')
    def test_webhook_handle_error(self, provider_class):
        provider_class.return_value.handle_webhook.side_effect = Exception('Error during handle_webhook')
        self.client.force_login(self.user)
        with self.assertRaisesRegex(Exception, 'Error during handle_webhook'):
            self.client.post(reverse('api:import-hook', kwargs={'pk': self.repo.id}))
        self.assertTrue(provider_class.return_value.handle_webhook.called)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to, ['user@user.fr'])
        self.assertEqual(mail.outbox[0].subject, 'An error occurred during Git hook handling')
        self.assertEqual(mail.outbox[0].body, MAIL_CONTENT)

    def test_webhook_handle_token_refresh_error(self):
        """
        The Git import hook should send an email for OAuthCredentials token refresh failures
        """
        self.client.force_login(self.user)
        self.creds.refresh_token = None
        self.creds.expiry = None
        self.assertTrue(self.creds.expired)
        self.creds.save()

        self.client.post(
            reverse('api:import-hook', kwargs={'pk': self.repo.id}),
            {
                'object_kind': 'push',
                'ref': 'refs/heads/something',
                'commits': [],
                'checkout_sha': 'abcd',
            },
            HTTP_X_GITLAB_EVENT='Push Hook',
            HTTP_X_GITLAB_TOKEN=self.repo.hook_token,
        )

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to, ['user@user.fr'])
        self.assertEqual(mail.outbox[0].subject, 'An error occurred during Git hook handling')
        self.assertEqual(mail.outbox[0].body, REFRESH_TOKEN_MAIL_CONTENT)

    @patch('arkindex.process.api.Repository.provider_class')
    def test_webhook(self, provider_class):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-hook', kwargs={'pk': self.repo.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertTrue(provider_class.return_value.handle_webhook.called)
        self.assertEqual(len(mail.outbox), 0)

    @patch('arkindex.process.models.Process.run')
    def test_start_imports_workers(self, run_mock):
        self.assertEqual(self.rev.processes.count(), 0)
        GitProvider.start_imports(None, rev=self.rev)
        self.assertEqual(self.rev.processes.count(), 1)
        self.assertEqual(run_mock.call_count, 1)
        _, run_kwargs = run_mock.call_args
        self.assertDictEqual(run_kwargs, {})
