from textwrap import dedent
from uuid import UUID

from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view
from rest_framework import permissions, serializers, status
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    GenericAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.response import Response

from arkindex.documents.models import Corpus
from arkindex.process.utils import annotate_image_url
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, TrainingModelMixin
from arkindex.project.pagination import CustomCursorPagination
from arkindex.project.permissions import IsVerified
from arkindex.training.models import (
    Dataset,
    DatasetElement,
    DatasetState,
    MetricValue,
    Model,
    ModelVersion,
    ModelVersionState,
)
from arkindex.training.serializers import (
    CreateModelErrorResponseSerializer,
    DatasetElementSerializer,
    DatasetLightSerializer,
    DatasetSerializer,
    MetricValueBulkSerializer,
    MetricValueCreateSerializer,
    ModelSerializer,
    ModelVersionSerializer,
    ModelVersionValidateSerializer,
    SelectionDatasetElementSerializer,
)
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        operation_id='ListModelVersions',
        description=(
            'List available versions of a Machine Learning model.\n\n'
            'Requires a **guest** access to the model, for versions that are available and have a tag,'
            'and a **contributor** access otherwise.'
        ),
    ),
    post=extend_schema(
        operation_id='CreateModelVersion',
        description=(
            'Create a new version for a Machine Learning model.\n\n'
            'The response includes an S3 URL that you can use to upload the archive containing the model.\n\n'
            'Once the archive is successfully uploaded, you can set the size and hash attributes using '
            '`PartialUpdateModelVersion`. The state of the model version will then be updated from `created` '
            'to `available`, meaning it can be used in a process.\n\n'
            'Requires a **contributor** access to the model.'
        ),
        responses={
            200: ModelVersionSerializer, 400: None, 403: None
        },
    ),
)
class ModelVersionsList(TrainingModelMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ModelVersionSerializer

    def get_queryset(self):
        self.model = self.get_object()
        qs = self.model.versions.order_by('-created')
        if self.access_level < Role.Contributor.value:
            qs = qs.filter(tag__isnull=False, state=ModelVersionState.Available)
        return qs

    def get_object(self):
        model = get_object_or_404(Model.objects.using('default').all(), id=self.kwargs['pk'])
        self.check_object_permissions(self.request, model)
        return model

    def check_object_permissions(self, request, model):
        self.access_level = get_max_level(self.request.user, model)
        needed_level, error_msg = (
            Role.Guest.value,
            'You need a Guest access to list versions of a model.',
        )
        if request.method not in permissions.SAFE_METHODS:
            needed_level, error_msg = (
                Role.Contributor.value,
                'You need a Contributor access to the model to create a new version.',
            )
        if not self.access_level or self.access_level < needed_level:
            raise PermissionDenied(detail=error_msg)
        return super().check_object_permissions(request, model)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context.update({
            'model': self.model,
            'is_contributor': self.access_level and self.access_level >= Role.Contributor.value,
        })
        return context

    def create(self, request, *args, **kwargs):
        self.model = self.get_object()
        return super().create(request, *args, **kwargs)


@extend_schema(tags=['training'])
@extend_schema_view(
    put=extend_schema(
        description=(
            'Update a version of a Machine Learning model.\n'
            'If the model version is updated with the hash of another available model version, '
            'an HTTP error 400 is returned containing data of the available model.\n\n'
            'Requires a **contributor** access to the model.'
        ),
    ),
    patch=extend_schema(
        description='Partially update a version of a Machine Learning model.\n\nRequires a **contributor** access to the model.'
    ),
    delete=extend_schema(
        description='Delete a model version.\n\nRequires an **admin** access on the related model.'
    )
)
class ModelVersionsRetrieve(TrainingModelMixin, RetrieveUpdateDestroyAPIView):
    """Retrieve a version of a Machine Learning model.

    Requires a **guest** access to the model, for versions that are available and have a tag,
    and a **contributor** access otherwise.
    """
    permission_classes = (IsVerified, )
    serializer_class = ModelVersionSerializer
    queryset = ModelVersion.objects.select_related("model")

    def get_serializer_context(self):
        context = super().get_serializer_context()
        access_level = getattr(self, 'access_level', None)
        context.update({
            'is_contributor': access_level is not None and access_level >= Role.Contributor.value,
        })
        return context

    def check_object_permissions(self, request, model_version):
        self.access_level = get_max_level(self.request.user, model_version.model)

        if request.method in permissions.SAFE_METHODS:
            # A Contributor access is required to retrieve an untagged or non available version
            if model_version.state != ModelVersionState.Available or model_version.tag is None:
                needed_level, error_msg = (
                    Role.Contributor.value,
                    'You need a Contributor access to the model to retrieve this version.',
                )
            else:
                needed_level, error_msg = (
                    Role.Guest.value,
                    'You need a Guest access to the model to retrieve this version.',
                )
        elif request.method.lower() == 'delete':
            needed_level, error_msg = (
                Role.Admin.value,
                'You need an Admin access to the model to destroy this version.',
            )
        else:
            needed_level, error_msg = (
                Role.Contributor.value,
                'You need a Contributor access to the model to update this version.',
            )

        if not self.access_level or self.access_level < needed_level:
            raise PermissionDenied(detail=error_msg)
        return super().check_object_permissions(request, model_version)


class ValidateModelVersion(TrainingModelMixin, GenericAPIView):
    """
    Checks the archive of a model version has correctly been uploaded to S3.
    If the verification succeeds, archive's attributes are updated and the state is set to `available`.
    If the verification fails, the state is set to `erroneous`.
    Once available, model versions cannot be validated anymore.
    If the hash conflicts with another existing model version, an HTTP error 409 is returned
    with the existing model version it the response body.

    Requires a **contributor** access to the model.
    """
    serializer_class = ModelVersionValidateSerializer
    queryset = ModelVersion.objects.select_related('model')

    def check_object_permissions(self, request, model_version):
        if not self.has_write_access(model_version.model):
            raise PermissionDenied(detail='You need a Contributor access to the model to validate this version.')
        if model_version.state == ModelVersionState.Available:
            raise PermissionDenied(detail='This model version is already marked as available.')
        return super().check_object_permissions(request, model_version)

    @extend_schema(
        operation_id='ValidateModelVersion',
        responses={
            200: ModelVersionSerializer,
            400: None,
            403: None,
            409: ModelVersionSerializer,
        },
        tags=['training'],
    )
    def post(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)

        # Ensure hash unicity among versions of a model
        existing_model_version = (
            instance.model.versions
            .filter(
                state=ModelVersionState.Available,
                hash=serializer.validated_data['hash'],
            ).first()
        )
        if existing_model_version:
            # Set the current model version as erroneous and return the available one
            instance.state = ModelVersionState.Error
            instance.save(update_fields=["state"])
            return Response(
                ModelVersionSerializer(
                    existing_model_version,
                    context={'is_contributor': True, 'model': instance},
                ).data,
                status=status.HTTP_409_CONFLICT,
            )

        for attr, value in serializer.validated_data.items():
            setattr(instance, attr, value)

        # Perform archive's validation
        try:
            instance.perform_check(save=False, raise_exc=True)
            instance.check_hash(save=False, raise_exc=True)
        except (AssertionError, ValueError) as e:
            # Set the model version as erroneous
            instance.save(update_fields=["state"])
            raise ValidationError({'detail': [str(e)]})

        instance.save()

        return Response(
            ModelVersionSerializer(instance).data,
            status=status.HTTP_201_CREATED,
        )


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        operation_id='ListModels',
        description=(
            'List available Machine Learning models.'
        ),
        responses={
            200: ModelSerializer,
        },
        parameters=[
            OpenApiParameter(
                'name',
                description='Filter models whose name contains the given string (case insensitive)',
                required=False,
            ),
            OpenApiParameter(
                'compatible_worker',
                description='Only return models that are compatible with this worker.',
                type=UUID,
                required=False,
            ),
        ]
    ),
    post=extend_schema(
        description=(
            'Create a new Machine Learning model.'
        ),
        responses={
            200: ModelSerializer,
            400: CreateModelErrorResponseSerializer,
            403: None
        },
    ),
)
class ModelsList(TrainingModelMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ModelSerializer

    def get_queryset(self):
        filters = Q()
        if 'name' in self.request.query_params:
            filters &= Q(name__icontains=self.request.query_params['name'])
        if 'compatible_worker' in self.request.query_params:
            data = self.request.query_params['compatible_worker']
            try:
                data = UUID(data)
            except (TypeError, ValueError):
                raise serializers.ValidationError({'compatible_worker': [f'{data} is not a valid UUID.']})
            filters &= Q(compatible_workers__id=data)

        # Use the default database to prevent a stale read when a model has just been created
        return self.readable_models.using('default').filter(filters).order_by('name')

    def perform_create(self, serializer):
        model_name = serializer.validated_data.get('name')
        existing_model = Model.objects.using('default').filter(name=model_name).first()
        if existing_model:
            if self.has_read_access(existing_model):
                raise ValidationError({
                    'id': str(existing_model.id),
                    'name': 'A model with this name already exists',
                })
            else:
                raise PermissionDenied()
        return serializer.save()


@extend_schema(tags=['training'])
class ModelRetrieve(TrainingModelMixin, RetrieveAPIView):
    """
    Retrieve a Machine Learning model.

    Requires a **guest** access to the model.
    """
    permission_classes = (IsVerified, )
    serializer_class = ModelSerializer

    def get_queryset(self):
        return self.readable_models


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        operation_id='DownloadModelVersion',
        description=(
            'Retrieve a download url for a model_version. A secret token is required to have access to the model version.'
            'This endpoint does **not** require to be logged in.'
        ),
        responses={
            302: serializers.Serializer,
        },
        parameters=[
            OpenApiParameter(
                'token',
                description='SHA256 hash used for authentication and check access to the model version.',
                required=True,
            ),
        ]
    ),
)
class ModelVersionDownload(TrainingModelMixin, RetrieveAPIView):
    queryset = ModelVersion.objects.all()

    def check_object_permissions(self, request, model_version):
        if model_version.state != ModelVersionState.Available:
            raise PermissionDenied(detail='This model version is not available yet.')
        token = model_version.build_authentication_token_hash()
        if token != request.query_params['token']:
            raise PermissionDenied(detail='Wrong token.')

    def get(self, request, *args, **kwargs):
        return Response(status=status.HTTP_302_FOUND, headers={'Location': self.get_object().s3_url})


@extend_schema(tags=['training'])
@extend_schema_view(
    post=extend_schema(
        operation_id='CreateMetricValue',
        description=(
            'Create a metric on a model version training.\n\n'
            'Requires a **contributor** access to the model.'
        ),
        responses={
            200: MetricValueCreateSerializer, 400: None, 403: None
        },
    )
)
class MetricValueCreate(CreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = MetricValueCreateSerializer
    queryset = MetricValue.objects.none()


@extend_schema(tags=['training'])
@extend_schema_view(
    post=extend_schema(
        operation_id='CreateMetricValues',
        description=(
            'Bulk create metrics on a model version training, optionally on a specified step.\n\n'
            'Requires a **contributor** access to the model.'
        ),
        responses={
            200: MetricValueCreateSerializer, 400: None, 403: None
        },
    )
)
class MetricValueBulkCreate(CreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = MetricValueBulkSerializer
    queryset = MetricValue.objects.none()


@extend_schema_view(
    get=extend_schema(
        operation_id='ListCorpusDatasets',
        description=dedent(
            """
            List all datasets in a corpus.

            Requires a **guest** access to the corpus.
            """
        ),
    ),
    post=extend_schema(
        operation_id='CreateDataset',
        description=dedent(
            """
            Create a new dataset in a corpus.

            Requires a **contributor** access to the corpus.
            """
        )
    ),
)
@extend_schema(tags=['datasets'])
class CorpusDataset(CorpusACLMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = DatasetLightSerializer
    queryset = Dataset.objects.none()

    def get_queryset(self):
        return Dataset.objects \
            .filter(corpus=self.corpus) \
            .order_by('name')

    @cached_property
    def corpus(self):
        role = Role.Guest
        if self.request.method not in permissions.SAFE_METHODS:
            role = Role.Contributor
        return self.get_corpus(self.kwargs['pk'], role=role)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Ignore this step when generating the schema with OpenAPI
        if not self.kwargs:
            return context
        context['corpus'] = self.corpus
        return context


@extend_schema_view(
    get=extend_schema(
        description=dedent(
            """
            Retrieve a dataset.

            Requires a **guest** access to the dataset's corpus.
            """
        )
    ),
    put=extend_schema(
        description=dedent(
            """
            Update a dataset.

            Requires a **contributor** access to the dataset's corpus. A dataset's state can only be updated by a Ponos task.
            """
        ),
    ),
    patch=extend_schema(
        description=dedent(
            """
            Partially update a dataset.

            Requires a **contributor** access to the dataset's corpus. A dataset's state can only be updated by a Ponos task.
            """
        )
    ),
    delete=extend_schema(
        description=dedent(
            """
            Delete a dataset. Only datasets in an `open` state can be deleted.

            Requires an **admin** access to the dataset's corpus.
            """
        )
    )
)
@extend_schema(tags=['datasets'])
class DatasetUpdate(ACLMixin, RetrieveUpdateDestroyAPIView):
    permission_classes = (IsVerified,)
    serializer_class = DatasetSerializer

    def get_queryset(self):
        queryset = Dataset.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))
        return queryset.select_related('corpus', 'creator')

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)

        if request.method in permissions.SAFE_METHODS:
            return
        role = Role.Contributor
        if request.method == 'DELETE':
            role = Role.Admin
            if obj.state != DatasetState.Open:
                raise PermissionDenied(detail='Only datasets in an open state can be deleted.')
        if not self.has_access(obj.corpus, role.value):
            raise PermissionDenied(detail=f'You do not have {str(role).lower()} access to corpus {obj.corpus.name}.')


@extend_schema(tags=['datasets'])
class DatasetElements(CorpusACLMixin, ListAPIView):
    """
    List all elements in a dataset.\n\n
    Requires a **guest** access to the dataset corpus.
    """
    permission_classes = (IsVerified, )
    queryset = DatasetElement.objects.none()
    serializer_class = DatasetElementSerializer

    @property
    def paginator(self):
        """
        Use a cursor pagination, ordering results by set then element id
        """
        if not hasattr(self, '_paginator'):
            self._paginator = CustomCursorPagination(ordering=('set', 'element_id'))
        return self._paginator

    def paginate_queryset(self, queryset):
        """
        Manually annotate the related elements with attributes for the nested serializer
        """
        page = super().paginate_queryset(queryset)
        for dataset_elt in page:
            dataset_elt.element.image_url = dataset_elt.image_url
            image = dataset_elt.element.image
            if image is not None:
                dataset_elt.element.image__width = image.width
                dataset_elt.element.image__height = image.height
        return page

    def get_queryset(self):
        dataset = get_object_or_404(
            Dataset.objects.select_related('corpus'),
            id=self.kwargs['pk'],
        )
        if not self.has_read_access(dataset.corpus):
            raise PermissionDenied(detail='You do not have access to the corpus of this dataset')
        return (
            dataset.dataset_elements
            .select_related('element__image')
            .annotate(image_url=annotate_image_url(field='element__image'))
            .only(
                'id',
                'set',
                'dataset_id',
                'element__name',
                'element__type_id',
                'element__confidence',
                'element__image_id',
                'element__image__width',
                'element__image__height',
                'element__polygon',
                'element__rotation_angle',
                'element__mirrored',
            )
        )


@extend_schema(tags=['datasets'])
@extend_schema_view(
    post=extend_schema(
        operation_id='CreateDatasetElementsSelection',
        responses={
            204: None,
            400: None,
            403: None,
        },
        parameters=[
            OpenApiParameter(
                'id',
                type=UUID,
                location=OpenApiParameter.PATH,
                description='ID of the corpus containing the selected elements.',
                required=True,
            )
        ],
    )
)
class CreateDatasetElementsSelection(CorpusACLMixin, CreateAPIView):
    """
    Add elements from a selection to a dataset.

    Requires a **contributor** access to the corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = SelectionDatasetElementSerializer

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)

    def check_object_permissions(self, request, corpus):
        if not self.has_write_access(corpus):
            raise PermissionDenied(detail="You need a Contributor access to the corpus to perform this action.")
        super().check_object_permissions(request, corpus)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['corpus'] = self.corpus
        return context

    @cached_property
    def corpus(self):
        return self.get_object()

    def create(self, request, *args, **kwargs):
        super().create(request, *args, **kwargs)
        return Response(status=status.HTTP_204_NO_CONTENT)
